function animalInit()
	animal = {}
	animal.body = FindBody("body",false)
	animal.state = "done"
	animal.upsideDownTimer = -1 -- -1 if not on head count up if on head
	animal.vel = Vec()
	animal.transform = Transform()
	animal.pos = Vec()
	animal.rot = Quat()
	animal.center = Vec()
	animal.fwd = Vec()
	animal.croak = 5
	animal.mass = GetBodyMass(animal.body)
	animal.prey = -1 -- this should be a body
	animal.allShapes = FindShapes("")
	animal.allBodies = FindBodies("")
	animal.legJoint = FindJoints("leg")
	if HasTag(animal.body,"pet") then animal.pet = true else animal.pet = false end

	ApplyBodyImpulse(animal.body,animal.center,Vec(0,1,0))

	for i=1, #animal.allBodies do 
		SetBodyActive(animal.allBodies[i],true)
	end
end

function navigationInit()
	navigation = {}
	navigation.state = "done"
	navigation.target = -1 -- is Vec if there is a target if not its -1 (is that allowed?)
	navigation.targetNode = VecCopy(animal.pos) -- Jump towards
	navigation.route = {} -- path nodes get stored here
	navigation.timer = -1 -- How long it took to get to the nearest path node. If navigation takes longer than 10 seconds it will recalculate
	navigation.jumpTimer = 1 -- delay between jumps
	navigation.wait = -1 -- a delay between moves
	navigation.blocked = false -- If blocked then animal wont move
	navigation.think = 0 -- How long query was searching
	navigation.attempt = 0
end

function attackInit()
	local loc = FindLocation("attack")
	attack = {}
	if loc ~= 0 then 
		attack.location = loc
		attack.type = GetTagValue(attack.location,"attack")
		attack.parentBody = FindBody(GetTagValue(attack.location,"parent"))
		local bt = GetBodyTransform(attack.parentBody)
		local lt = GetLocationTransform(attack.location)
		attack.localTransform = TransformToLocalTransform(bt,lt)
		attack.timer = 0 
		attack.state = "done"
		if attack.type == "tounge" then 
			attack.toungeLenght = 0
			attack.toungeGrabbed = 0 
		end
	end
end

function init()
	hop = LoadSound("MOD/hop.ogg")
	deth = LoadSound("MOD/deth.ogg")
	croak = LoadSound("MOD/croak.ogg")

	Time = 0

	sndD = false
	sndH = false

	boing = false

	animalInit() -- Has To be before all of them
	navigationInit()
	attackInit()

	toungeSprite = LoadSprite("gfx/laser.png")
end

function hunt(dt)
	if attack.type == "tounge" and animal.prey ~= 0 then
		local flyPos = GetBodyTransform(animal.prey).pos
		queryRejectAllShapes(animal.allShapes)
		local dir = VecNormalize(VecSub(flyPos,animal.pos))
		local dist = VecDist(flyPos,animal.pos)
		local hit, _,_, shape = QueryRaycast(animal.pos,dir,dist,0,false)
		local bool = (hit == false or HasTag(GetShapeBody(shape),"fly"))
		if bool and VecDot(animal.fwd,dir) > 0.5 and dist < 10 or attack.toungeGrabbed == true then
			navigation.blocked = true
			attack.state = "aim"
			local btParent= GetBodyTransform(attack.parentBody)
			local tParent= TransformToParentTransform(btParent, attack.localTransform)
			toungePos = VecLerp(tParent.pos,flyPos,math.clamp(attack.toungeLenght/dist,0,1))

			attack.toungeLenght = attack.toungeLenght + VecDist(toungePos,animal.pos)/4+0.1

			local length = VecDist(tParent.pos,toungePos)
			local t = Transform(VecLerp(tParent.pos,toungePos, 0.5))
			local xAxis = VecNormalize(VecSub(tParent.pos, toungePos))
			local zAxis = VecNormalize(VecSub(tParent.pos, GetCameraTransform().pos))
			t.rot = QuatAlignXZ(xAxis, zAxis)
			DrawSprite(toungeSprite, t, length, 0.08, 0.713, 0.435, 0.313, 10, true)

			if math.clamp(attack.toungeLenght/VecDist(toungePos,animal.pos),0,1) == 1 then 
				attack.toungeGrabbed = true 
				ConstrainPosition(animal.prey,0,flyPos,animal.pos,40,40)
				if dist < 1 then 
					Delete(animal.prey)
					animal.prey = 0
					attack.toungeGrabbed = false
					animal.state = "done"
					attack.state = "done"
					attack.toungeLenght = 0
					PlaySound(croak,animal.pos,0.5)
					navigation.blocked = false
				end
			end
		else 
			attack.toungeGrabbed = false
			attack.toungeLenght = 0
			attack.state = "done"
			navigation.blocked = false
		end
	end
end

function animalMovement(dt)
	debugAnimal()
	animalNavigation(dt)
	navigationTurn(dt)
	targetUpdate(dt)
	hunt(dt)

	for i=1, #animal.allBodies do
		if animal.pos[2] < -20 then 
			tpFromVoid()
			break
		end
	end
	
end

function targetUpdate(dt)

	if #navigation.route == 0 and navigation.state == "move" and animal.inWater == false then 
		navigation.state = "done"
	end

	local flies = FindBodies("fly",true)
    local distance = 0
	local cloestFly = 0
    for i=1, #flies do
		local flyPos = GetBodyTransform(flies[i]).pos
        if i == 1 then 
            distance = VecDist(flyPos,animal.pos)
            cloestFly = flies[i]
        else
            if distance > VecDist(flyPos,animal.pos) then 
				distance = VecDist(flyPos,animal.pos)
                cloestFly = flies[i]
            end
        end
    end

	local cloesestFlypos = VecCopy(GetBodyTransform(cloestFly).pos)
	local bool = 15 > VecDist(cloesestFlypos,animal.pos) and cloestFly ~= 0


	if (bool and animal.state ~= "hunt") or (bool and animal.state == "hunt" and VecStr(navigation.target) == VecStr(Vec())) then 
		navigation.route = {}
		navigation.state = "find"
		animal.state = "hunt"
		navigation.target = VecCopy(cloesestFlypos)
		navigation.blocked = false
		animal.prey = cloestFly
		
	elseif animal.state == "hunt" and 16 < VecDist(cloesestFlypos,animal.pos) or animal.state == "hunt" and #flies == 0 then
		animal.state = "done"
		animal.prey = 0
		navigation.state = "done"
		navigation.blocked = false
	end

	if animal.state == "hunt" and navigation.state ~= "find" then
		local pos = GetBodyTransform(animal.prey).pos
		if VecDist(navigation.target,pos) > 3 then 
			navigation.state = "find"
			navigation.target = VecCopy(pos)
			navigation.blocked = false
		end
	end
	
	if navigation.attempt >= 3 then 
		navigationReset()
	end


	if 16 < VecDist(cloesestFlypos,animal.pos) and navigation.state == "done" then
		if animal.pet == false then 
			local newPos = VecAdd(animal.pos,Vec(math.random(-40,40),30,math.random(-40,40)))
			local hit, dist = QueryRaycast(newPos,Vec(0,-1,0),30,0.5,false)
			if hit then
				navigation.target = VecAdd(newPos,VecScale(Vec(0,-1,0),dist))
				navigation.state = "find"
				navigation.blocked = false
			end
		elseif animal.pet == true then
			navigation.target = GetPlayerTransform().pos
			navigation.state = "find"
			navigation.blocked = false
		end
	end
end

function debugAnimal()
	

	local pt = GetPlayerCameraTransform()
	local hit, dist ,normal, shape = QueryRaycast(pt.pos,TransformToParentVec(pt,Vec(0,0,-1)),20,0,true)

	if InputPressed("q") and hit and animal.body == GetShapeBody(shape) then
			if HasTag(animal.body,"check") then 
				RemoveTag(animal.body,'check')
			else
				SetTag(animal.body,"check")
			end
		end
		if HasTag(animal.body,"check") then
		DrawBodyOutline(animal.body,1,1,1,1)
		
		DebugLine(animal.pos,navigation.target,1,1,1,1)

		DebugWatch("Transform",animal.transform)
		DebugWatch("Velocity",animal.vel)
		DebugWatch("Navigation State",navigation.state)
		DebugWatch("Navigation Progress",GetPathState())
		DebugWatch("navigation jump",navigation.jumpTimer)
		DebugWatch("lenght",#navigation.route)
		DebugWatch("animal on head", animal.upsideDown)
		DebugWatch("nav.think", navigation.think)
		DebugWatch("animal state", animal.state)
		DebugWatch("navigation.target",navigation.target)
		DebugWatch("is Pet", animal.pet)
		DebugWatch("navigation.blocked",navigation.blocked)
		
		if InputDown("c") then 
			DebugPrint("")
		end

		for i=1, #navigation.route-1 do
			DebugLine(navigation.route[i], navigation.route[i+1])
			DebugCross(navigation.route[i],1,0,1,1)
		end

		local playert = GetPlayerCameraTransform()
		local dir = TransformToParentVec(playert,Vec(0,0,-1))
		local hit, dist = QueryRaycast(playert.pos,dir,50,0,false)
		if hit and InputPressed("t") then 
			navigation.target = VecAdd(playert.pos,VecScale(dir,dist))
			navigation.route = {}
			navigation.state = "find"
		end
	end
end

function isDebug()
	if HasTag(animal.body,"check") then 
		return true
	end
	return false 
end

function navigationReset()
	navigation.state = "done"
	navigation.attempt = 0
	navigation.jumpTimer = 0 
	navigation.route = {}
	navigation.think = 0 
	navigation.blocked = false
end

function animalNavigation(dt)

	if navigation.state == "find" and navigation.blocked == false then 
	
		if navigation.target ~= -1 then
		
			queryRejectAllShapes(animal.allShapes)
			--SetInt("level.frog.parallelism",GetInt("level.frog.parallelism")+1)
			QueryPath(animal.pos,navigation.target,100,0)
			navigation.state = "search"
			navigation.think = 0
		end
	end
	local state = GetPathState()

	if state == "busy" then 
		navigation.think = navigation.think + dt 
	end

	if navigation.state == "search" and state == "idle" then 
		navigation.state = "done" 
	end

	local bool = (state == "done" or state == "fail" or navigation.think > 5)
	if bool and navigation.state == "search" then
		AbortPath()
		--SetInt("level.frog.parallelism",GetInt("level.frog.parallelism")-1)
		navigation.state = "move"
		navigation.route = {}
		navigation.think = 0
		local length=GetPathLength()
		local l=0
		while l < length do
			navigation.route[#navigation.route+1] = GetPathPoint(l)
			l = l + 0.5
		end
	end


	-- Prune walked on path --

	if #navigation.route ~= 0 then
		local closestPoint = #navigation.route
		for num,point in ipairs(navigation.route) do
			if VecDist(navigation.route[closestPoint],animal.pos) > VecDist(point,animal.pos) then 
				closestPoint = num
			end
		end
		--(navigation.timer .. " df")

		local offset = 0
		for i=1, closestPoint-1 do 
			table.remove(navigation.route,i+offset)
			offset = offset+1
			navigation.timer = 0
		end
	end

	local bool = navigation.state == "move" and animal.upsideDown == false
	if bool and #navigation.route ~= 0 or bool and animal.inWater == true then
		if navigation.timer > 10 then 
			navigation.state = "find"
			navigation.attempt = navigation.attempt + 1
			navigation.timer = 0
		else 
			navigation.timer = navigation.timer + dt
		end

		local jumpForce = math.clamp(VecSub(navigation.targetNode,navigation.route[1])[2]/0.1,1,2)
		local jumpVec = VecAdd(Vec(0,700*jumpForce,0),TransformToParentVec(animal.transform,Vec(0,0,-400*jumpForce)))

		for i=2, 10 do 
			local itter = 10-i 

			local dirLeft = QuatRotateVec(QuatEuler(8,i*i,0),Vec(0,0,-1))
			local dirRight = QuatRotateVec(QuatEuler(8,(i*i)*-1,0),Vec(0,0,-1))
			local rayDist = math.clamp(0.7+(2/itter),0,1.1)
			queryRejectAllShapes(animal.allShapes)

			local hit, dist, normal, shape = QueryRaycast(animal.center,TransformToParentVec(animal.transform,dirLeft),rayDist,0,false)
			
			if hit == false then
				if isDebug() then 
					DebugLine(animal.center,VecAdd(animal.center,VecScale(TransformToParentVec(animal.transform,dirLeft),dist)),1,0,1,1)
				end
			else 
				if isDebug() then 
					DebugLine(animal.center,VecAdd(animal.center,VecScale(TransformToParentVec(animal.transform,dirLeft),dist)),1,0,0,1)
				end
				
				jumpVec = QuatRotateVec(QuatEuler(0,(itter*itter)*-1,0),jumpVec)
			end

			queryRejectAllShapes(animal.allShapes)
			local hit, dist, normal, shape = QueryRaycast(animal.center,TransformToParentVec(animal.transform,dirRight),rayDist,0,false)

			if hit == false then
				if isDebug() then 
					DebugLine(animal.center,VecAdd(animal.center,VecScale(TransformToParentVec(animal.transform,dirRight),dist)))
				end
				
			else 
				if isDebug() then 
					DebugLine(animal.center,VecAdd(animal.center,VecScale(TransformToParentVec(animal.transform,dirRight),dist)),1,0,0,1)
				end

				jumpVec = QuatRotateVec(QuatEuler(0,(itter*itter),0),jumpVec)
			end
		end

		--DebugLine(animal.pos,VecAdd(animal.pos,VecScale(jumpVec,0.001)))
		if navigation.jumpTimer < 0 and animal.onGround and attack.state ~= "aim" then 
			local ahead = math.clamp(VecDot(animal.fwd,VecNormalize(VecSub(navigation.targetNode,animal.pos))),0.5,1)
			if animal.state ~= "hunt" then
				navigation.jumpTimer = math.random()+math.random(0,2)
				if animal.pet == true then 
					jumpVec = VecAdd(jumpVec,TransformToParentVec(animal.transform,Vec(0,0,-300*ahead)))
				end
			else 
				jumpVec = VecAdd(jumpVec,TransformToParentVec(animal.transform,Vec(0,0,-300*ahead)))
				navigation.jumpTimer = math.random()
			end

			ApplyBodyImpulse(animal.body,animal.center,jumpVec)

			for i=1, #animal.legJoint do 
				SetJointMotor(animal.legJoint[i],0,0)
			end

			navigation.state = "jump"
			navigation.blocked = true
			navigation.wait = 2

			PlaySound(hop,animal.sound,0.5)
			SetBodyAngularVelocity(animal.body,TransformToParentVec(animal.transform,Vec(1,0,0)))
		else 
			navigation.jumpTimer = navigation.jumpTimer -dt
		end

		local bool = navigation.state == "move" and #navigation.route ~= 0
		if VecDist(animal.pos,navigation.route[#navigation.route]) < 3 and bool or navigation.state == "move" and VecDist(animal.pos,navigation.target) < 3 then 
			navigationReset()
		end

		if 15 < VecDist(animal.pos,navigation.route[closestPoint]) and VecStr(navigation.route[closestPoint]) ~= VecStr(Vec()) then 
			navigation.state = "find"
		end
		--DebugLine(animal.pos,navigation.route[closestPoint],1,0,1,1)
	end

	if navigation.state == "jump" then 
		if 8 < animal.gForce then 
			SetBodyVelocity(animal.body,VecScale(animal.vel,0.5))
			navigation.state = "move"
			navigation.blocked = false
			for i=1, #animal.legJoint do 
				SetJointMotorTarget(animal.legJoint[i],0,10)
			end
		end
		if navigation.wait < 0 then 
			navigation.state = "move"
			navigation.blocked = false
			for i=1, #animal.legJoint do 
				SetJointMotorTarget(animal.legJoint[i],0,10)
			end
		else
			navigation.wait = navigation.wait - dt
		end
	end

	if animal.swimming > 0 then 
		ApplyBodyImpulse(animal.body,animal.center,Vec(0,(animal.mass*7*dt)))
	end

	if animal.upsideDown == true and animal.onGround == false and navigation.state ~= "jump" then
		ApplyBodyImpulse(animal.body,animal.center,Vec(0,(animal.mass*9.5*dt)))
	end

end

function navigationTurn()
	-- Air Controll --
	queryRejectAllShapes(animal.allShapes)
	local hit, dist, normal, shape = QueryRaycast(animal.center,VecNormalize(animal.vel),10,0,false)
	--if hit and VecLength(animal.vel) > 0.05 then 

	local target = navigation.route[1]
	if 6 < #navigation.route then 
		for i=1, 6 do 
			target = VecLerp(target,navigation.route[i],0.5)
		end
	else 
		for i=1, #navigation.route do 
			target = VecLerp(target,navigation.route[i],0.5)
		end
	end

	if animal.swimming > 0 then 
		target = navigation.target
	end

	navigation.targetNode = VecCopy(target)

	local rot = TransformToLocalTransform(animal.transform,QuatLookAt(animal.pos,QuatRotateVec(TransformToParentVec(animal.transform,normal),QuatEuler(0,0,90))))
	local localX,localY,localZ = GetQuatEuler(rot)
	local parentRot = TransformToParentTransform(animal.transform,QuatEuler(localX,localY,-50))

	local dir = QuatLookAt(animal.pos,target)					
	local _,dirYaw,_ = GetQuatEuler(dir)  


	local parentRotRoll,_,parentRotPitch = GetQuatEuler(parentRot)

	rot = QuatEuler(parentRotRoll,dirYaw,parentRotPitch)


	local coef = 0.3/dist+1
	ConstrainOrientation(animal.body,0,animal.rot,rot,20,coef+2)
	--end
end

function animalUpdate(dt)
	animal.gForce = (VecLength(animal.vel) - VecLength(GetBodyVelocity(animal.body))) / (dt * 10) -- G-Force of last tick
	animal.vel = GetBodyVelocity(animal.body)
	animal.transform = GetBodyTransform(animal.body)
	animal.pos = animal.transform.pos
	animal.rot = animal.transform.rot

	local com = GetBodyCenterOfMass(animal.body)
	animal.center = TransformToParentPoint(animal.transform,com)
	animal.fwd = TransformToParentVec(animal.transform,Vec(0,0,-1))
	animal.onGround = isAnimalOn("ground",0.8)
	if IsPointInWater(animal.pos) then 
		animal.onGround = true 
		animal.inWater = true
	else 
		animal.inWater = false
	end

	animal.upsideDown = isAnimalOn("head",1,true)

	if animal.croak < 0 then 
		animal.croak = math.random(3,10)
		PlaySound(croak,animal.pos,0.1)
	else
		animal.croak  = animal.croak  - dt
	end
	local bool, depth = IsPointInWater(animal.pos)
	animal.swimming = depth

	if animal.onGround == false and animal.upsideDown == false then 
		animal.upsideDownTimer = animal.upsideDownTimer + dt
	elseif animal.upsideDown == true and animal.onGround == false then 
		animal.upsideDownTimer = animal.upsideDownTimer + dt*2
	else 
		animal.upsideDownTimer = 0 
	end

	if animal.upsideDownTimer > 5 then 
		local x,y,z = GetQuatEuler(animal.rot)
		ConstrainOrientation(animal.body,0,animal.rot,QuatEuler(0,y,0),2)
	end
end


function isAnimalOn(on,lenght, rejectGrass)
	local dir = Vec()
	if not rejectGrass then rejectGrass = false end
	if on == "head" then 
		dir = VecCopy(Vec(0,1,0))
	elseif on == "ground" then 
		dir = VecCopy(Vec(0,-1,0))
	end
	queryRejectAllShapes(animal.allShapes)
	QueryRequire("physical")
	local hit, dist, normal, shape = QueryRaycast(animal.center,TransformToParentVec(animal.transform,dir),lenght,0,false)
	if rejectGrass == true and GetShapeMaterialAtPosition(shape,VecAdd(animal.center,VecScale(dist,dir))) == "" and hit == true then 
		return false 
	end
	return hit
end


function update(dt)
	if animal.state ~= "dead" then
		animalUpdate(dt)
		animalMovement(dt)
		if IsBodyBroken(animal.body) then 
			animal.state = "dead"
			PlaySound(deth,animal.pos,0.1)
		end
	elseif not removed then 
		removed = true
		RemoveTag(animal.body,"frog")
		
	end
end

function VecDist(a, b)
	return VecLength(VecSub(a, b))
end

function math.clamp(val, lower, upper)
    if lower > upper then lower, upper = upper, lower end -- swap if boundaries supplied the wrong way
    return math.max(lower, math.min(upper, val))
end

function queryRejectAllShapes(shapes)
	for i=1, #shapes do 
		QueryRejectShape(shapes[i])
	end
end


function tpFromVoid()
	for j=1, 300 do 
		local pos = Vec(math.random(-400,400),200,math.random(-400,400))
		local hit, dist = QueryRaycast(pos,Vec(0,-1,0),300,0,false)
		if hit then 
			for i=1,#animal.allBodies do 
				SetBodyTransform(animal.allBodies[i],Transform(VecAdd(pos,VecScale(Vec(0,-1,0),dist))))
				SetBodyVelocity(animal.allBodies[i],Vec(0,0,0))
			end
			return
		end
	end
	animal.state = "dead"
end